package moca.core.proxy;

import java.net.InetSocketAddress;
import java.util.HashMap;
import java.util.Map;

import moca.core.proxy.communication.CommunicationProtocol;
import moca.core.proxy.message.Message;
import moca.core.proxy.receivers.CISMessage;
import moca.core.proxy.receivers.Package;
import moca.service.cis.util.DeviceContext;

/**
* Gerencia as informaes dos clientes conectados ao proxy. As informaes de cada cliente
* esto associadas a um identificador nico, com o qual podem ser recuperadas.
* 
* @author hana
*/

public class ClientManager {
	/**
	* Mapa que associa identificador do cliente s suas informaes
	*/
	private Map clientMap = new HashMap();
	
	/** Mapa reverso que associa o InetAddress do cliente a seu identificador */
	private Map addressMap = new HashMap();

	private static final ClientManager me = new ClientManager() ;
	/**
	* Registra um cliente
	* @param clientInfo informaoes do cliente a ser registrado
	*/
	
	public final synchronized void register(ClientInfo clientInfo) {
		clientMap.put(clientInfo.getClientId(), clientInfo);
		if(clientInfo.getCommunicationProtocol().getCommunicMode().equals(CommunicationProtocol.SYNC) && clientInfo.getCommunicationProtocol().getProtocol().equals(CommunicationProtocol.TCP))// && clientInfo.getCommunicationProtocol().getProtocol().equals(CommunicationProtocol.TCP))
			addressMap.put(clientInfo.getCommunicationProtocol().getSyncTCPConnection().getRemoteAddress(), clientInfo.getClientId());
		else
			addressMap.put(clientInfo.getCommunicationProtocol().getAddress(), clientInfo.getClientId());
	}
	/**
	* Desregistra um cliente
	* @param clientId identificador do cliente
	*/
	public final synchronized ClientInfo unregister(String clientId) {
		ClientInfo cf = (ClientInfo) clientMap.remove(clientId);
		addressMap.remove(cf.getCommunicationProtocol().getAddress());
		return cf;
	}
	
	private ClientManager(){
		// construtor private para evitar q a classe seja instanciada com new.
	}
	
	public static ClientManager getInstance(){
		return me;
	}
	/**
	* Desregistra todos os cliente 
	*/
	public final synchronized void unregisterAll() {
		clientMap.clear();
		addressMap.clear();
	}
	/**
	 * Recupera as informaes do cliente mediante seu identificador
	 * @param clientId identificador unico do cliente
	 * @return informaes do cliente
	 */
	public final synchronized ClientInfo getClientInfo(String clientId) {
		return (ClientInfo) clientMap.get(clientId);
	}
	/**
	 * Recupera o identificador do cliente mediante seu endereo de rede
	 * @param clientAddress Endereo do cliente
	 * @return identificador do cliente
	 */
	public final synchronized String getClientId(InetSocketAddress clientAddress) {
		return (String) addressMap.get(clientAddress);
	}	
	/**
	 * Valida a mensagem fornecida, envolvendo-a com um pacote interno do proxy
	 * @param clientId identificador do cliente
	 * @param msg mensagem
	 * @return pacote que envolve a mensagem, que contem informaes de controle para
	 * uso interno ao proxy
	 */
	public final synchronized Package validate(String clientId, Message msg) {
		ClientInfo info = this.getClientInfo(clientId);
		if (info != null) {
			Package pack = new Package(info, msg);
			pack.setPreviousPackage(info.getLastPackage());
			info.setLastPackage(pack);
			return pack;
		}
		return null;
	}
	/**
	 * Valida uma mensagem CIS
	 * @param clientId identificador interno
	 * @param msg mensagem CIS
	 * @param currentContext Contexto corrente do dispositivo
	 * @return pacote CIS
	 */
	public final synchronized Package validateCIS(
		                            String clientId,
		                            CISMessage msg,
		                            DeviceContext currentContext) {
		ClientInfo info = this.getClientInfo(clientId);
		if (info != null) {
            //Qdo o evento  da expresso ALL_DEVICES (tag para sempre executar uma adaptao),
			// o contexto  nulo e ento mantm o contexto anterior
			if(currentContext != null){
				info.setCurrentContext(currentContext);
			}
			Package pack = new Package(info, msg);
			return pack;
		}
		return null;
	}
}
