package moca.core.proxy.message;

import java.io.Serializable;
import java.net.InetSocketAddress;


/**
 *  Definition of the message accepted by the proxy. 
 *  The message must contain some information (like adressee) that the proxy needs to 
 *  intermediate the communication between aplication's server and clients, 
 *  and perform the adaptations when nescessary.   
 *  <p>
 *   To use the proxy features, it is highly recommended to use this class or an extension of it.  
 * 
 *  @author hana 
 *  
 */
public class DefaultMessage implements Message, Serializable{

	/**
	 * 
	 */
	private static final long serialVersionUID = 2263440618700652608L;

	/** The sender of the message */
	private String sender;
		
	//TODO  no caso de tpico precisa de eventProperties tb?? Qdo eu oferecer opo de 
	// expressions no PUB/SUB
	/** To whom the message must be sent. 
	 *  In the case of a mobile client, the identification is its MAC Address.
	 */
	private String addressee;
	
	/**  Subject to which the message will be published. */
	private String subject;
	
	/** Message type: DATA or CONTROL */
	private int msgType;
	
	private InetSocketAddress tcpSyncClientsocket = null ;
	/**
	 * Specify the kind of data is been sent in the message. 
	 * Data type is in MIME-like format, or a special tag: OBJECT.
	 * Messages that have OBJECT data type are not adapt by the proxy 
	 * (they are only forwarded, or cached)
	 */
	private String dataType;
	
	/** Data that must be sent to the addressee, and can be adapted */
	private byte[] data;
	
	
	/** Default constructor - creates an empty message */
	public DefaultMessage(){
		this.sender = null;
		this.addressee = null;
		this.msgType = DATA;
		this.dataType = OBJECT;
		this.data = null;
		this.subject = null;
		
	}
	
	/**
	 * Constructor
	 * 
	 * @param sender  - the sender of the message
	 * @param addressee - the addressee (synchronous communication)
	 * @param msgType - message type: Message.DATA or Message.CONTROL
	 * @param dataType - data type in mime format string or Message.OBJECT
	 * @param data - data of the message
	 */
	public DefaultMessage(String sender, String addressee, int msgType,
			String dataType, byte[] data) {
		this.sender = sender;
		this.addressee = addressee;
		this.msgType = msgType;
		this.dataType = dataType;
		this.data = data;
		this.subject = null;
	}
	
	
	/**
	 * Constructor
	 * 
	 * @param sender  - the sender of the message
	 * @param subject - the subject of interest (asynchronous communication)
	 * @param msgType - message type: Message.DATA or Message.CONTROL
	 * @param dataType - data type in mime format string or Message.OBJECT
	 * @param data - data of the message
	 * @param async - Indicate the asynchronous communication (only to differ from the other constructor)
	 */
	public DefaultMessage(String sender, String subject, int msgType,
			String dataType, byte[] data, boolean async) {
		this.sender = sender;
		this.subject = subject;
		this.msgType = msgType;
		this.dataType = dataType;
		this.data = data;
		this.addressee = null;
	}
	
	/**
	 * @return Returns the addressee.
	 */
	public String getAddressee() {
		return addressee;
	}
	/**
	 * @param addressee The addressee to set.
	 */
	public void setAddressee(String addressee) {
		this.addressee = addressee;
	}
	/**
	 * @return Returns the data.
	 */
	public byte[] getData() {
		return data;
	}
	/**
	 * @param data The data to set.
	 */
	public void setData(byte[] data) {
		this.data = data;
	}
	/**
	 * @return Returns the sender.
	 */
	public String getSender() {
		return sender;
	}
	/**
	 * @param sender The sender to set.
	 */
	public void setSender(String sender) {
		this.sender = sender;
	}
	/**
	 * @param dataType The dataType to set.
	 */
	public void setDataType(String dataType) {
		this.dataType = dataType;
	}
	/**
	 * @param msgType The msgType to set.
	 */
	public void setMsgType(int msgType) {
		this.msgType = msgType;
	}
	
	/**
	 * @return Returns the dataType. This type is 
	 *  a string similar to the MIME format, or a special tag OBJECT ="object"
	 */
	public String getDataType() {
		return dataType;
	}
	/**
	 * @return Returns the msgType - Message.DATA or Message.CONTROL.
	 */
	public int getMsgType() {
		return msgType;
	}
	
	/**
	 * @return Returns the subject.
	 */
	public String getSubject() {
		return subject;
	}
	/**
	 * @param subject The subject to set.
	 */
	public void setSubject(String subject) {
		this.subject = subject;
	}
	
	/**
	 *  Clone the object
	 */
	public Object clone() throws CloneNotSupportedException {
		DefaultMessage msg = (DefaultMessage) super.clone();
		if (msg.data != null) {
			msg.data = (byte[]) msg.data.clone();	
		}
		return msg;

	}

	public void setClientAddress(InetSocketAddress socket) {
		tcpSyncClientsocket = socket ;
		
	}

	public InetSocketAddress getClientAddress() {
       return tcpSyncClientsocket;
	}
}