package moca.core.proxy.state;

import java.util.Iterator;

import moca.core.proxy.ClientInfo;
import moca.core.proxy.actions.adapters.AdaptationException;
import moca.core.proxy.actions.adapters.Adapter;
import moca.core.proxy.message.Message;

/**
 * Armazena o estado de um determinado usuario. Este objeto  atualizado
 * de acordo com as informaes fornecidas pelo CIS
 * @author hana
 */
public class State {
	/**
	* Estado de configurao (cadastrado no arquivo de configuracoes)
	*/
	private StateConf stateConf;
	/**
	 * Situao do cliente com relacao ao estado (est ou no esta no estado)
	 */
	private boolean isOn;
	/**
	 * Adapter do estado. Agrega todos os adapters cadastrados para este estado
	 */
	private Adapter stateAdapter;
	/**
	 * Construtor. O estado  criado a partir de um estado de configuracao
	 * @param stateConf estado de configuracao
	 */
	public State(StateConf stateConf) {
		this.stateConf = stateConf;
	}
	
	/**
	 * Executa os listeners associados ao estado.
	 * Este mtodo ser executado sempre que houver uma mudana na situao 
	 * monitorada do cliente
	 * @param isOn indica o tipo de transicao 
	 * 			(off para on -> true)
	 * 			(on para off -> false) 
	 * @param disp objeto despachante
	 * @param clientInfo objeto com informacoes sobre o cliente
	 */
	public void runListeners(
		boolean isOn,
		moca.core.proxy.receivers.Dispatcher disp,
		ClientInfo clientInfo) {
		this.stateConf.runListeners(isOn, disp, clientInfo);
	}
	/**
	* Seta o estado do cliente (on ou off) e retorna o estado anterior
	* @param isOn estado do cliente
	* @return true se o estado anterior era on e false caso contrario
	*/
	private boolean setIsOn(boolean isOn) {
		if (this.isOn != isOn) {
			this.isOn = isOn;
			return !isOn;
		}
		return this.isOn;
	}
	/**
	* Esta classe agrega todos os adapters cadastrados para este estado
	* Ela executa cada um dos adapters, caso o estado esteja on e no 
	* os executa se estiver off
	* @author hana
	*/
	public class StateAdapter extends Adapter {
		public Message execute(ClientInfo clientInfo, Message msg)
			throws AdaptationException {
			if (State.this.isOn) { 
				Iterator it = stateConf.getRuleSet().iterator();
				while (it.hasNext()) {
					Rule r = (Rule) it.next();
					Adapter adapter = r.getAdapter();
					msg = adapter.execute(clientInfo, msg);
					if (msg == null) {
						break;
					}
				}
			}
			return msg;
		}
	}
	/**
	 * Seta o estado para on
	 * @return retorna o estado anterior
	 */
	public synchronized boolean setOn() {
		return this.setIsOn(true);
	}
	/**
	 * Seta o estado para off
	 * @return retorna o estado anterior
	 */
	public synchronized boolean setOff() {
		return this.setIsOn(false);
	}
	
	/**
	 * Verifica se o estado est ativo
	 * @return o status do estado
	 */
	public boolean isOn(){
		return isOn;
	}
	
	/**
	 * Retorna o adapter do estado. O adapter do estado agrega todos os adapters
	 * configurados para este estado.
	 * @return adapter do estado
	 */
	public synchronized Adapter getAdapter() {
		if (this.stateAdapter == null) {
			this.stateAdapter = new StateAdapter();
		}
		return this.stateAdapter;
	}

	/**
	 * @return Returns the stateConf.
	 */
	public StateConf getStateConf() {
		return stateConf;
	}
}
